// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Leonardo Mezzina, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using BoPi.Common;
using System.Diagnostics;

namespace BoPi.Types
{
  public class TypesOperations
  {
		public static int GetHashCode(IType t)
		{
			Debug.Assert(t != null);
			if (t is Bind) {
				Debug.Assert(false);
				return GetHashCode(((Bind) t).Content);
			} else if (t is Bottom) {
				return 0;
			} else if (t is Union) {
				Union t1 = (Union) t;
				return GetHashCode(t1.Fst) ^ GetHashCode(t1.Snd);
			} else if (t is Void) {
				return 1;
			} else if (t is ConstantTypeName) {
				return ((ConstantTypeName) t).Name.GetHashCode();
			} else if (t is Sequence) {
				Sequence t1 = (Sequence) t;
				return GetHashCode(t1.Top) ^ GetHashCode(t1.Tail);
			} else if (t is Labelled) {
				Labelled t1 = (Labelled) t;
				return t1.Labels.GetHashCode() ^ GetHashCode(t1.Content);
			} else if (t is IntLiteral) {
				return ((IntLiteral) t).Val.GetHashCode();
			} else if (t is IntType) {
				return 2;
			} else if (t is StringLiteral) {
				return ((StringLiteral) t).Val.GetHashCode();
			} else if (t is StringType) {
				return 3;
			} else if (t is Chan) {
				Chan t1 = (Chan) t;
				return t1.Capability.GetHashCode() ^ GetHashCode(t1.Carried);
			} else if (t is ErrorType) {
				return 4;
			} else {
				Debug.Assert(false);
				return -1;
			}
		}

		public static bool Equals(IType s, IType t)
		{
			if (Object.ReferenceEquals(s, t)) {
				return true;
			} else if (s is Bind && t is Bind) {
				Debug.Assert(false);
				return (Equals(((Bind) s).Content, ((Bind) t).Content));
			} else if (s is Bottom && t is Bottom) {
				return true;
			} else if (s is Union && t is Union) {
				Union s1 = (Union) s;
				Union t1 = (Union) t;
				return Equals(s1.Fst, t1.Fst) && Equals(s1.Snd, t1.Snd);
			} else if (s is Void && t is Void) {
				return true;
			} else if (s is ConstantTypeName && t is ConstantTypeName) {
				ConstantTypeName s1 = (ConstantTypeName) s;
				ConstantTypeName t1 = (ConstantTypeName) t;
				return (Object.ReferenceEquals(s1.Entry, t1.Entry));
			} else if (s is Sequence && t is Sequence) {
				Sequence s1 = (Sequence) s;
				Sequence t1 = (Sequence) t;
				return Equals(s1.Top, t1.Top) && Equals(s1.Tail, t1.Tail);
			} else if (s is Labelled && t is Labelled) {
				Labelled s1 = (Labelled) s;
				Labelled t1 = (Labelled) t;
				return s1.Labels.Includes(t1.Labels) && t1.Labels.Includes(s1.Labels) && Equals(s1.Content, t1.Content);
			} else if (s is IntLiteral && t is IntLiteral) {
				return (((IntLiteral) s).Val == ((IntLiteral) t).Val);
			} else if (s is IntType && t is IntType) {
				return true;
			} else if (s is StringLiteral && t is StringLiteral) {
				return (((StringLiteral) s).Val == ((StringLiteral) t).Val);
			} else if (s is StringType && t is StringType) {
				return true;
			} else if (s is Chan && t is Chan) {
				Chan s1 = (Chan) s;
				Chan t1 = (Chan) t;
				return s1.Capability == t1.Capability && Equals(s1.Carried, t1.Carried);
			} else if (s is ErrorType && t is ErrorType) {
				return true;
			} else
				return false;
		}

    public static IType Union(IType s, IType t)
    { return new Union(s, t); }

    /// <summary>
    /// Computes the Intersection between two types.
    /// </summary>
    /// <param name="s">the fist type</param>
    /// <param name="t">the second type</param>
    /// <returns>the intersect type</returns>
    public static bool EmptyIntersection(IType s, IType t)
    {
      IType intersection = Intersect(s, t, new Dictionary<string, ISymbolTableTypeEntry>());
      if (intersection.IsErrorType()) 
        return false;
      else if ((new EmptynessChecker()).IsEmpty(intersection)) 
        return true;
      else return false;
    }

    /// <summary>
    /// Computes the Intersection between two types.
    /// </summary>
    /// <param name="s">the fist type</param>
    /// <param name="t">the second type</param>
    /// <returns>the intersect type</returns>
    /// <remarks>the method is internal because it is not precise: ErrorType is used when the intersection 
    /// cannot be expressed as schema.
    /// </remarks>
    internal static IType Intersect(IType s, IType t)
    { return Intersect(s, t, new Dictionary<string, ISymbolTableTypeEntry>()); }

    /// <summary>
    /// Computes the Intersection between two types.
    /// </summary>
    /// <param name="s">the fist type</param>
    /// <param name="t">the second type</param>
    /// <param name="computing">pairs of type names the algorithm is calculating</param>
    /// <returns>the intersect type</returns>
    private static IType Intersect(IType s, IType t, IDictionary<string, ISymbolTableTypeEntry> inC)
    {
      SubtypingChecker sc = new SubtypingChecker();
      if (s.IsBottom() || t.IsBottom())				
        return s;
      if (s.IsErrorType() || t.IsErrorType())
        return s;
      if (s.IsUnion()) 
      { 
        IType int1 = Intersect(s.AsUnion().Fst, t, inC);
        IType int2 = Intersect(s.AsUnion().Snd, t, inC);
        if (int1.IsBottom() && int2.IsBottom()) return new Bottom();
        else if (int1.IsBottom()) return int2;
        else if (int2.IsBottom()) return int1;
        else return new Union(int1,int2);
      }
      if (t.IsUnion()) return Intersect(t, s, inC);
      if (s.IsConstantTypeName() && t.IsConstantTypeName()) 
      {
        //these names are unique because of "#". 
        string name1 = s.AsConstantTypeName() + "#" + t.AsConstantTypeName();
        string name2 = t.AsConstantTypeName() + "#" + s.AsConstantTypeName();
        if (!inC.ContainsKey(name1) && !inC.ContainsKey(name2)) //we start to compute the Intersection
        {
          IType defofs = s.AsConstantTypeName().Entry.Type;
          IType defoft = t.AsConstantTypeName().Entry.Type;
          if (defofs.IsErrorType() || defoft.IsErrorType()) return new ErrorType();
					ISymbolTableTypeEntry entry1 = new SymbolTableTypeEntry(name1, null);
					inC.Add(name1, entry1);
					ISymbolTableTypeEntry entry2 = null;
					// LUCA: if name1 == name2 then one entry is enough (the hash table would complain otherwise)
					if (name1 != name2) {
						entry2 = new SymbolTableTypeEntry(name2, null);
						inC.Add(name2, entry2);
					}
          IType int1= Intersect(defofs, defoft, inC);
					entry1.Type = int1;
					if (name1 != name2)
						entry2.Type = int1;
          return new ConstantTypeName(name1, entry1);
        }
        //we already have the Intersection
        else if (inC.ContainsKey(name1)) return new ConstantTypeName(name1, (ISymbolTableTypeEntry) inC[name1]);
        else return new ConstantTypeName(name2, (ISymbolTableTypeEntry) inC[name2]);
      }
      else if (s.IsConstantTypeName()) 
        return Intersect(s.AsConstantTypeName().Entry.Type, t, inC);
      else if (t.IsConstantTypeName()) 
        return Intersect(s, t.AsConstantTypeName().Entry.Type, inC);
      else if (s.IsBaseType() || s.IsVoid()) 
      {
        if (sc.IsSubtype(s, t)) return s;
        else if (sc.IsSubtype(t, s)) return t; 
        else return new Bottom();
      }
      else if (s.IsSequence())
      {
        if (t.IsSequence())
        {
          IType top = Intersect(s.AsSequence().Top, t.AsSequence().Top, inC);
          if (top.IsBottom()) return new Bottom();
          IType tail = Intersect(s.AsSequence().Tail, t.AsSequence().Tail, inC);
          if (tail.IsBottom()) return new Bottom();
          return new Sequence(top.AsLabelled(), tail);
        }
        else if (t.IsLabelled())
          return Intersect(t, s, inC);
        else return new Bottom();
      }
      else if (s.IsChan()) 
      {
        if (t.IsChan())
        {
          IType cs = s.AsChan().Carried;
          IType ct = t.AsChan().Carried;
          // i && i = i, i && io = i, o && o = o, o && io = o, io && io = io, i && o = io
          if (s.AsChan().Capability == Chan.CAPABILITY.IN && t.AsChan().Capability != Chan.CAPABILITY.OUT)
            return new Chan(Intersect(cs, ct, inC), Chan.CAPABILITY.IN);
          else if (s.AsChan().Capability == Chan.CAPABILITY.INOUT && t.AsChan().Capability == Chan.CAPABILITY.IN) 
            return new Chan(Intersect(cs, ct, inC), Chan.CAPABILITY.IN);
          else if (s.AsChan().Capability == Chan.CAPABILITY.OUT && t.AsChan().Capability != Chan.CAPABILITY.IN)
            return new Chan(new Union(s.AsChan().Carried, t.AsChan().Carried), Chan.CAPABILITY.OUT);
            //return new Chan(DeterminedUnion(new Union(cs, ct)), Chan.CAPABILITY.OUT);
          else if (s.AsChan().Capability == Chan.CAPABILITY.INOUT && t.AsChan().Capability == Chan.CAPABILITY.OUT)
            return new Chan(new Union(s.AsChan().Carried, t.AsChan().Carried), Chan.CAPABILITY.OUT);
            //return new Chan(DeterminedUnion(new Union(cs, ct)), Chan.CAPABILITY.OUT);
          else if (s.AsChan().Capability == Chan.CAPABILITY.INOUT && t.AsChan().Capability == Chan.CAPABILITY.INOUT) 
          { 
            if (sc.IsSubtype(cs, ct) && sc.IsSubtype(ct, cs))
              return s;
            else return new Bottom();
          }
          if (s.AsChan().Capability == Chan.CAPABILITY.OUT && t.AsChan().Capability == Chan.CAPABILITY.IN)
            return Intersect(t, s, inC);
          else if (s.AsChan().Capability == Chan.CAPABILITY.IN && t.AsChan().Capability == Chan.CAPABILITY.OUT)
          {
            if (sc.IsSubtype(cs, ct) && sc.IsSubtype(ct,cs))
              return new Chan(cs, Chan.CAPABILITY.INOUT);
            else if (!sc.IsSubtype(ct, cs))
              return new Bottom();
            else if (sc.IsSubtype(ct, cs))
              return new ErrorType(); //the intersection is not empty but for the moment i'm not able to compute it! SAMUELE:TODO
          }
        }
        else return new Bottom();
      }
      else if (s.IsLabelled()) 
      {
        if (t.IsLabelled())
        {
          ILabelsSet labels = LabelsSet.Intersection(s.AsLabelled().Labels, t.AsLabelled().Labels);
          if (!labels.IsEmpty())
          {
            IType content = Intersect(s.AsLabelled().Content, t.AsLabelled().Content, inC);
            return new Labelled((LabelsSet) labels, content);
          }
          else return new Bottom();
        } 
        else if (t.IsSequence()) 
        {
          if (sc.IsSubtype(new Void(), t.AsSequence().Tail)) 
            return TypesOperations.Intersect(s,t.AsSequence().Top, inC);
        }
        else return new Bottom();
      }
      throw new ApplicationException("private static Type intersect: Unhandled case " + s + " x " + t);
    }

    internal static IList<Sequence> Expand(IType t) 
    {
      IList<Sequence> exp = new List<Sequence>();
      Expand(t, exp, new List<IType>());
      return exp;
    }
    /// <summary>
    /// Given a type <c>t</c> returns the sequences it contains
    /// </summary>
    /// <param name="s">the sequence</param>
    /// <param name="t">the type to Expand</param>
    /// <returns>a set containing either guarded sequences or void obtained by Expanding <c>t</c></returns>	
    /// <remarks>constant type name must be guarded</remarks>
    private static IList<Sequence> Expand(IType t, IList<Sequence> exp, ICollection<IType> visited)
    {	
      if (t.IsUnion())
      {
        Expand(t.AsUnion().Fst, exp, visited);
        Expand(t.AsUnion().Snd, exp, visited);
      }
      else if (t.IsConstantTypeName() && !visited.Contains(t))
      {
        visited.Add(t);
        Expand(t.AsConstantTypeName().Entry.Type, exp, visited);
      }
      else if (t.IsSequence())
      {
        if (!exp.Contains(t.AsSequence())) exp.Add(t.AsSequence());
      }
      else if (t.IsLabelled())
      { 
        Sequence s = new Sequence(t.AsLabelled(), new Void());
        if (!exp.Contains(s)) exp.Add(s);
      }
      return exp;
    }

    internal static IType Unify(IList<IType> types)
    {
      if (types.Count == 0) return new Bottom();
      else if (types.Count == 1) return types[0];
      else 
      {
        IList<IType> inTypes = new List<IType>();
        IType u = types[0];
        inTypes.Add(types[0]);
        for (int i = 1; i < types.Count; i++)
        {
          if (!inTypes.Contains(types[i]))
          {
            inTypes.Add(types[i]);
            u = new Union(u, types[i]);
          }
        }
        return u;
      }
    }
  }
}
